@@ -2,22 +2,22 @@
  * gawkapi.c -- Implement the functions defined for gawkapi.h
  */
 
-/* 
- * Copyright (C) 2012, 2013 the Free Software Foundation, Inc.
- * 
+/*
+ * Copyright (C) 2012-2019, 2021, 2022, 2023, the Free Software Foundation, Inc.
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -25,7 +25,14 @@
 
 #include "awk.h"
 
+/* Declare some globals used by api_get_file: */
+extern IOBUF *curfile;
+extern INSTRUCTION *main_beginfile;
+extern int currule;
+
 static awk_bool_t node_to_awk_value(NODE *node, awk_value_t *result, awk_valtype_t wanted);
+static const char *valtype2str(awk_valtype_t type);
+static NODE *ns_lookup(const char *name_space, const char *name, char **full_name);
 
 /*
  * api_get_argument --- get the count'th paramater, zero-based.
@@ -43,7 +50,7 @@ api_get_argument(awk_ext_id_t id, size_t count,
 	NODE *arg;
 
 	if (result == NULL)
-		return false;
+		return awk_false;
 
 	(void) id;
 
@@ -59,44 +66,44 @@ api_get_argument(awk_ext_id_t id, size_t count,
 	 */
 	arg = get_argument(count);
 	if (arg == NULL)
-		return false;
+		return awk_false;
 
 	/* if type is undefined */
-	if (arg->type == Node_var_new) {
+	if (arg->type == Node_var_new || arg->type == Node_elem_new) {
 		if (wanted == AWK_UNDEFINED)
-			return true;
+			return awk_true;
 		else if (wanted == AWK_ARRAY) {
 			goto array;
 		} else {
 			goto scalar;
 		}
 	}
-	
+
 	/* at this point, we have real type */
 	if (arg->type == Node_var_array || arg->type == Node_array_ref) {
 		if (wanted != AWK_ARRAY && wanted != AWK_UNDEFINED)
-			return false;
+			return awk_false;
 		goto array;
 	} else
 		goto scalar;
 
 array:
 	/* get the array here */
-	arg = get_array_argument(count, false);
+	arg = get_array_argument(arg, count);
 	if (arg == NULL)
-		return false;
+		return awk_false;
 
 	return node_to_awk_value(arg, result, wanted);
 
 scalar:
 	/* at this point we have a real type that is not an array */
-	arg = get_scalar_argument(count, false);
+	arg = get_scalar_argument(arg, count);
 	if (arg == NULL)
-		return false;
+		return awk_false;
 
 	return node_to_awk_value(arg, result, wanted);
 #else
-	return false;
+	return awk_false;
 #endif
 }
 
@@ -114,23 +121,23 @@ api_set_argument(awk_ext_id_t id,
 	(void) id;
 
 	if (array == NULL || array->type != Node_var_array)
-		return false;
+		return awk_false;
 
 	if (   (arg = get_argument(count)) == NULL
-	    || arg->type != Node_var_new)
-		return false;
+	    || (arg->type != Node_var_new && arg->type != Node_elem_new))
+		return awk_false;
 
-	arg = get_array_argument(count, false);
+	arg = get_array_argument(arg, count);
 	if (arg == NULL)
-		return false;
+		return awk_false;
 
 	array->vname = arg->vname;
 	*arg = *array;
 	freenode(array);
 
-	return true;
+	return awk_true;
 #else
-	return false;
+	return awk_false;
 #endif
 }
 
@@ -139,8 +146,11 @@ api_set_argument(awk_ext_id_t id,
 NODE *
 awk_value_to_node(const awk_value_t *retval)
 {
-	NODE *ext_ret_val;
+	NODE *ext_ret_val = NULL;
 	NODE *v;
+#ifdef HAVE_MPFR
+	int tval = 0;
+#endif
 
 	if (retval == NULL)
 		fatal(_("awk_value_to_node: received null retval"));
@@ -152,13 +162,57 @@ awk_value_to_node(const awk_value_t *retval)
 	case AWK_UNDEFINED:
 		ext_ret_val = dupnode(Nnull_string);
 		break;
+	case AWK_BOOL:
+		ext_ret_val = make_bool_node(retval->bool_value != awk_false);
+		break;
 	case AWK_NUMBER:
-		ext_ret_val = make_number(retval->num_value);
+		switch (retval->num_type) {
+		case AWK_NUMBER_TYPE_DOUBLE:
+			ext_ret_val = make_number(retval->num_value);
+			break;
+		case AWK_NUMBER_TYPE_MPFR:
+#ifdef HAVE_MPFR
+			if (! do_mpfr)
+				fatal(_("awk_value_to_node: not in MPFR mode"));
+			ext_ret_val = make_number_node(MPFN);
+			mpfr_init(ext_ret_val->mpg_numbr);
+			tval = mpfr_set(ext_ret_val->mpg_numbr, (mpfr_srcptr) retval->num_ptr, ROUND_MODE);
+			IEEE_FMT(ext_ret_val->mpg_numbr, tval);
+			mpfr_clear(retval->num_ptr);
+#else
+			fatal(_("awk_value_to_node: MPFR not supported"));
+#endif
+			break;
+		case AWK_NUMBER_TYPE_MPZ:
+#ifdef HAVE_MPFR
+			if (! do_mpfr)
+				fatal(_("awk_value_to_node: not in MPFR mode"));
+			ext_ret_val = make_number_node(MPZN);
+			mpz_init(ext_ret_val->mpg_i);
+			mpz_set(ext_ret_val->mpg_i, (mpz_ptr) retval->num_ptr);
+			mpz_clear(retval->num_ptr);
+#else
+			fatal(_("awk_value_to_node: MPFR not supported"));
+#endif
+			break;
+		default:
+			fatal(_("awk_value_to_node: invalid number type `%d'"), retval->num_type);
+			break;
+		}
 		break;
 	case AWK_STRING:
 		ext_ret_val = make_str_node(retval->str_value.str,
 				retval->str_value.len, ALREADY_MALLOCED);
 		break;
+	case AWK_STRNUM:
+		ext_ret_val = make_str_node(retval->str_value.str,
+				retval->str_value.len, ALREADY_MALLOCED);
+		ext_ret_val->flags |= USER_INPUT;
+		break;
+	case AWK_REGEX:
+		ext_ret_val = make_typed_regex(retval->str_value.str,
+				retval->str_value.len);
+		break;
 	case AWK_SCALAR:
 		v = (NODE *) retval->scalar_cookie;
 		if (v->type != Node_var)
@@ -193,7 +247,21 @@ api_fatal(awk_ext_id_t id, const char *format, ...)
 	va_end(args);
 }
 
-/* api_warning --- print a warning message and exit */
+/* api_nonfatal --- print a non fatal error message */
+
+static void
+api_nonfatal(awk_ext_id_t id, const char *format, ...)
+{
+	va_list args;
+
+	(void) id;
+
+	va_start(args, format);
+	err(false, _("error: "), format, args);
+	va_end(args);
+}
+
+/* api_warning --- print a warning message */
 
 static void
 api_warning(awk_ext_id_t id, const char *format, ...)
@@ -217,13 +285,12 @@ api_lintwarn(awk_ext_id_t id, const char *format, ...)
 	(void) id;
 
 	va_start(args, format);
-	if (lintwarn == r_fatal) {
+	if (lintfunc == r_fatal) {
 		err(true, _("fatal: "), format, args);
-		va_end(args);
 	} else {
 		err(false, _("warning: "), format, args);
-		va_end(args);
 	}
+	va_end(args);
 }
 
 /* api_register_input_parser --- register an input_parser; for opening files read-only */
@@ -239,7 +306,7 @@ api_register_input_parser(awk_ext_id_t id, awk_input_parser_t *input_parser)
 	register_input_parser(input_parser);
 }
 
-/* api_register_output_wrapper --- egister an output wrapper, for writing files / two-way pipes */
+/* api_register_output_wrapper --- register an output wrapper, for writing files / two-way pipes */
 
 static void api_register_output_wrapper(awk_ext_id_t id,
 		awk_output_wrapper_t *output_wrapper)
@@ -307,19 +374,21 @@ api_unset_ERRNO(awk_ext_id_t id)
 
 static awk_bool_t
 api_add_ext_func(awk_ext_id_t id,
-		const char *namespace,
-		const awk_ext_func_t *func)
+		const char *name_space,
+		awk_ext_func_t *func)
 {
 	(void) id;
-	(void) namespace;
 
 	if (func == NULL)
-		return false;
+		return awk_false;
+
+	if (name_space == NULL)
+		fatal(_("add_ext_func: received NULL name_space parameter"));
 
 #ifdef DYNAMIC
-	return make_builtin(func);
+	return make_builtin(name_space, func);
 #else
-	return false;
+	return awk_false;
 #endif
 }
 
@@ -373,12 +442,121 @@ api_awk_atexit(awk_ext_id_t id,
 	list_head = p;
 }
 
+static struct {
+	char **strings;
+	size_t i, size;
+} scopy;
+
+/* free_api_string_copies --- release memory used by string copies */
+
+void
+free_api_string_copies()
+{
+	size_t i;
+
+	for (i = 0; i < scopy.i; i++)
+		free(scopy.strings[i]);
+	scopy.i = 0;
+}
+
+/* assign_string --- return a string node with NUL termination */
+
+static inline void
+assign_string(NODE *node, awk_value_t *val, awk_valtype_t val_type)
+{
+	val->val_type = val_type;
+	if (node->stptr[node->stlen] != '\0') {
+		/*
+		 * This is an unterminated field string, so make a copy.
+		 * This should happen only for $n where n > 0 and n < NF.
+		 */
+		char *s;
+
+		assert((node->flags & MALLOC) == 0);
+		if (scopy.i == scopy.size) {
+			/* expand list */
+			if (scopy.size == 0)
+				scopy.size = 8;	/* initial size */
+			else
+				scopy.size *= 2;
+			erealloc(scopy.strings, char **, scopy.size * sizeof(char *), "assign_string");
+		}
+		emalloc(s, char *, node->stlen + 1, "assign_string");
+		memcpy(s, node->stptr, node->stlen);
+		s[node->stlen] = '\0';
+		val->str_value.str = scopy.strings[scopy.i++] = s;
+	}
+	else
+		val->str_value.str = node->stptr;
+	val->str_value.len = node->stlen;
+}
+
+/* assign_number -- return a number node */
+
+#define assign_double(val) \
+	val->num_value = node->numbr; \
+	val->num_type = AWK_NUMBER_TYPE_DOUBLE; \
+	val->num_ptr = NULL
+
+static inline void
+assign_number(NODE *node, awk_value_t *val)
+{
+	val->val_type = AWK_NUMBER;
+
+#ifndef HAVE_MPFR
+	assign_double(val);
+#else
+	switch (node->flags & (MPFN|MPZN)) {
+	case 0:
+		assign_double(val);
+		break;
+	case MPFN:
+		val->num_value = mpfr_get_d(node->mpg_numbr, ROUND_MODE);
+		val->num_type = AWK_NUMBER_TYPE_MPFR;
+		val->num_ptr = &node->mpg_numbr;
+		break;
+	case MPZN:
+		val->num_value = mpz_get_d(node->mpg_i);
+		val->num_type = AWK_NUMBER_TYPE_MPZ;
+		val->num_ptr = &node->mpg_i;
+		break;
+	default:
+		fatal(_("node_to_awk_value: detected invalid numeric flags combination `%s'; please file a bug report"), flags2str(node->flags));
+		break;
+	}
+#endif
+}
+#undef assign_double
+
+/* assign_regex --- return a regex node */
+
+static inline void
+assign_regex(NODE *node, awk_value_t *val)
+{
+	/* a REGEX node cannot be an unterminated field string */
+	assert((node->flags & MALLOC) != 0);
+	assert(node->stptr[node->stlen] == '\0');
+	val->str_value.str = node->stptr;
+	val->str_value.len = node->stlen;
+	val->val_type = AWK_REGEX;
+}
+
+/* assign_bool --- return a bool node */
+
+static inline void
+assign_bool(NODE *node, awk_value_t *val)
+{
+	assert((node->flags & BOOLVAL) != 0);
+	val->val_type = AWK_BOOL;
+	val->bool_value = get_number_si(node) != 0 ? awk_true : awk_false;
+}
+
 /* node_to_awk_value --- convert a node into a value for an extension */
 
 static awk_bool_t
 node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 {
-	awk_bool_t ret = false;
+	awk_bool_t ret = awk_false;
 
 	if (node == NULL)
 		fatal(_("node_to_awk_value: received null node"));
@@ -388,9 +566,10 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 	switch (node->type) {
 	case Node_var_new:	/* undefined variable */
+	case Node_elem_new:	/* undefined element */
 		val->val_type = AWK_UNDEFINED;
 		if (wanted == AWK_UNDEFINED) {
-			ret = true;
+			ret = awk_true;
 		}
 		break;
 
@@ -399,7 +578,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 		if (wanted == AWK_SCALAR) {
 			val->val_type = AWK_SCALAR;
 			val->scalar_cookie = (void *) node;
-			ret = true;
+			ret = awk_true;
 			break;
 		}
 
@@ -408,50 +587,157 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 	case Node_val:
 		/* a scalar value */
 		switch (wanted) {
-		case AWK_NUMBER:
-			val->val_type = AWK_NUMBER;
+		case AWK_BOOL:
+			if ((node->flags & BOOLVAL) != 0) {
+				assign_bool(node, val);
+				ret = awk_true;
+			} else
+				ret = awk_false;
+			break;
 
-			(void) force_number(node);
-			if ((node->flags & NUMCUR) != 0) {
-				val->num_value = get_number_d(node);
-				ret = true;
+		case AWK_NUMBER:
+			if ((node->flags & REGEX) != 0)
+				val->val_type = AWK_REGEX;
+			else {
+				(void) force_number(node);
+				assign_number(node, val);
+				ret = awk_true;
+			}
+			break;
+
+		case AWK_STRNUM:
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
+			case STRING:
+				val->val_type = AWK_STRING;
+				break;
+			case NUMBER:
+				(void) force_string(node);
+				/* fall through */
+			case NUMBER|USER_INPUT:
+				assign_string(node, val, AWK_STRNUM);
+				ret = awk_true;
+				break;
+			case REGEX:
+				val->val_type = AWK_REGEX;
+				break;
+			case NUMBER|STRING:
+				if (node == Nnull_string) {
+					val->val_type = AWK_UNDEFINED;
+					break;
+				}
+				/* fall through */
+			default:
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
+				val->val_type = AWK_UNDEFINED;
+				break;
 			}
 			break;
 
 		case AWK_STRING:
-			val->val_type = AWK_STRING;
-
 			(void) force_string(node);
-			if ((node->flags & STRCUR) != 0) {
-				val->str_value.str = node->stptr;
-				val->str_value.len = node->stlen;
-				ret = true;
+			assign_string(node, val, AWK_STRING);
+			ret = awk_true;
+			break;
+
+		case AWK_REGEX:
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case STRING:
+				val->val_type = AWK_STRING;
+				break;
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
+			case NUMBER:
+				val->val_type = AWK_NUMBER;
+				break;
+			case NUMBER|USER_INPUT:
+				val->val_type = AWK_STRNUM;
+				break;
+			case REGEX:
+				assign_regex(node, val);
+				ret = awk_true;
+				break;
+			case NUMBER|STRING:
+				if (node == Nnull_string) {
+					val->val_type = AWK_UNDEFINED;
+					break;
+				}
+				/* fall through */
+			default:
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
+				val->val_type = AWK_UNDEFINED;
+				break;
 			}
 			break;
 
 		case AWK_SCALAR:
-			if ((node->flags & NUMBER) != 0) {
-				val->val_type = AWK_NUMBER;
-			} else if ((node->flags & STRING) != 0) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
+			case STRING:
 				val->val_type = AWK_STRING;
-			} else
+				break;
+			case NUMBER:
+				val->val_type = AWK_NUMBER;
+				break;
+			case NUMBER|USER_INPUT:
+				val->val_type = AWK_STRNUM;
+				break;
+			case REGEX:
+				val->val_type = AWK_REGEX;
+				break;
+			case NUMBER|STRING:
+				if (node == Nnull_string) {
+					val->val_type = AWK_UNDEFINED;
+					break;
+				}
+				/* fall through */
+			default:
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
-			ret = false;
+				break;
+			}
 			break;
 
 		case AWK_UNDEFINED:
 			/* return true and actual type for request of undefined */
-			if ((node->flags & NUMBER) != 0) {
-				val->val_type = AWK_NUMBER;
-				val->num_value = get_number_d(node);
-				ret = true;
-			} else if ((node->flags & STRING) != 0) {
-				val->val_type = AWK_STRING;
-				val->str_value.str = node->stptr;
-				val->str_value.len = node->stlen;
-				ret = true;
-			} else
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				assign_bool(node, val);
+				ret = awk_true;
+				break;
+			case STRING:
+				assign_string(node, val, AWK_STRING);
+				ret = awk_true;
+				break;
+			case NUMBER:
+				assign_number(node, val);
+				ret = awk_true;
+				break;
+			case NUMBER|USER_INPUT:
+				assign_string(node, val, AWK_STRNUM);
+				ret = awk_true;
+				break;
+			case REGEX:
+				assign_regex(node, val);
+				ret = awk_true;
+				break;
+			case NUMBER|STRING:
+				if (node == Nnull_string) {
+					val->val_type = AWK_UNDEFINED;
+					ret = awk_true;
+					break;
+				}
+				/* fall through */
+			default:
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
+				break;
+			}
 			break;
 
 		case AWK_ARRAY:
@@ -464,15 +750,14 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 		val->val_type = AWK_ARRAY;
 		if (wanted == AWK_ARRAY || wanted == AWK_UNDEFINED) {
 			val->array_cookie = node;
-			ret = true;
-		} else {
-			ret = false;
-		}
+			ret = awk_true;
+		} else
+			ret = awk_false;
 		break;
 
 	default:
 		val->val_type = AWK_UNDEFINED;
-		ret = false;
+		ret = awk_false;
 		break;
 	}
 
@@ -484,7 +769,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
  * 	- No access to special variables (NF, etc.)
  * 	- One special exception: PROCINFO.
  *	- Use sym_update() to change a value, including from UNDEFINED
- *	  to scalar or array. 
+ *	  to scalar or array.
  */
 /*
  * Lookup a variable, fills in value. No messing with the value
@@ -498,6 +783,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 static awk_bool_t
 api_sym_lookup(awk_ext_id_t id,
+		const char *name_space,
 		const char *name,
 		awk_valtype_t wanted,
 		awk_value_t *result)
@@ -509,8 +795,13 @@ api_sym_lookup(awk_ext_id_t id,
 	if (   name == NULL
 	    || *name == '\0'
 	    || result == NULL
-	    || (node = lookup(name)) == NULL)
-		return false;
+	    || ! is_valid_identifier(name)
+	    || name_space == NULL
+	    || (name_space[0] != '\0' && ! is_valid_identifier(name_space)))
+		return awk_false;
+
+	if ((node = ns_lookup(name_space, name, NULL)) == NULL)
+		return awk_false;
 
 	if (is_off_limits_var(name))	/* a built-in variable */
 		node->flags |= NO_EXT_SET;
@@ -531,7 +822,7 @@ api_sym_lookup_scalar(awk_ext_id_t id,
 	if (node == NULL
 	    || result == NULL
 	    || node->type != Node_var)
-		return false;
+		return awk_false;
 
 	update_global_values();	/* make sure stuff like NF, NR, are up to date */
 
@@ -542,6 +833,7 @@ api_sym_lookup_scalar(awk_ext_id_t id,
 
 static awk_bool_t
 api_sym_update(awk_ext_id_t id,
+		const char *name_space,
 		const char *name,
 		awk_value_t *value)
 {
@@ -550,12 +842,17 @@ api_sym_update(awk_ext_id_t id,
 
 	if (   name == NULL
 	    || *name == '\0'
-	    || value == NULL)
-		return false;
+	    || value == NULL
+	    || ! is_valid_identifier(name)
+	    || name_space == NULL
+	    || (name_space[0] != '\0' && ! is_valid_identifier(name_space)))
+		return awk_false;
 
 	switch (value->val_type) {
 	case AWK_NUMBER:
+	case AWK_STRNUM:
 	case AWK_STRING:
+	case AWK_REGEX:
 	case AWK_UNDEFINED:
 	case AWK_ARRAY:
 	case AWK_SCALAR:
@@ -564,52 +861,68 @@ api_sym_update(awk_ext_id_t id,
 
 	default:
 		/* fatal(_("api_sym_update: invalid value for type of new value (%d)"), value->val_type); */
-		return false;
+		return awk_false;
 	}
 
-	node = lookup(name);
+	char *full_name = NULL;
+	node = ns_lookup(name_space, name, & full_name);
 
 	if (node == NULL) {
 		/* new value to be installed */
 		if (value->val_type == AWK_ARRAY) {
 			array_node = awk_value_to_node(value);
-			node = install_symbol(estrdup((char *) name, strlen(name)),
-					Node_var_array);
+			node = install_symbol(full_name, Node_var_array);
 			array_node->vname = node->vname;
 			*node = *array_node;
 			freenode(array_node);
 			value->array_cookie = node;	/* pass new cookie back to extension */
 		} else {
 			/* regular variable */
-			node = install_symbol(estrdup((char *) name, strlen(name)),
-					Node_var);
+			node = install_symbol(full_name, Node_var);
 			node->var_value = awk_value_to_node(value);
 		}
 
-		return true;
+		return awk_true;
 	}
 
 	/*
 	 * If we get here, then it exists already.  Any valid type is
-	 * OK except for AWK_ARRAY.
+	 * OK except for AWK_ARRAY (unless it is in Node_var_new undefined
+	 * state, in which case an array is OK).
 	 */
 	if (   (node->flags & NO_EXT_SET) != 0
-	    || is_off_limits_var(name)) {	/* most built-in vars not allowed */
+	    || is_off_limits_var(full_name)) {	/* most built-in vars not allowed */
 		node->flags |= NO_EXT_SET;
-		return false;
+		efree((void *) full_name);
+		return awk_false;
 	}
 
-	if (    value->val_type != AWK_ARRAY
-	    && (node->type == Node_var || node->type == Node_var_new)) {
+	efree((void *) full_name);
+
+	if (value->val_type == AWK_ARRAY) {
+		if (node->type == Node_var_new) {
+			/* special gymnastics to convert untyped to an array */
+			array_node = awk_value_to_node(value);
+			array_node->vname = node->vname;
+			unref(node->var_value);
+			*node = *array_node;
+			freenode(array_node);
+			value->array_cookie = node;	/* pass new cookie back to extension */
+			return awk_true;
+		}
+	} else if (node->type == Node_var
+		   || node->type == Node_var_new
+		   || node->type == Node_elem_new) {
 		unref(node->var_value);
 		node->var_value = awk_value_to_node(value);
-		if (node->type == Node_var_new && value->val_type != AWK_UNDEFINED)
+		if ((node->type == Node_var_new || node->type == Node_elem_new)
+		    && value->val_type != AWK_UNDEFINED)
 			node->type = Node_var;
 
-		return true;
+		return awk_true;
 	}
 
-	return false;
+	return awk_false;
 }
 
 /* api_sym_update_scalar --- update a scalar cookie */
@@ -625,7 +938,7 @@ api_sym_update_scalar(awk_ext_id_t id,
 	    || node == NULL
 	    || node->type != Node_var
 	    || (node->flags & NO_EXT_SET) != 0)
-		return false;
+		return awk_false;
 
 	/*
 	 * Optimization: if valref is 1, and the new value is a string or
@@ -650,10 +963,12 @@ api_sym_update_scalar(awk_ext_id_t id,
 			r->flags = MALLOC|NUMBER|NUMCUR;
 			r->stptr = NULL;
 			r->stlen = 0;
-			return true;
+			return awk_true;
 		}
 		break;
+
 	case AWK_STRING:
+	case AWK_STRNUM:
 		if (node->var_value->valref == 1) {
 			NODE *r = node->var_value;
 
@@ -667,25 +982,33 @@ api_sym_update_scalar(awk_ext_id_t id,
 			/* make_str_node(s, l, ALREADY_MALLOCED): */
 			r->numbr = 0;
 			r->flags = (MALLOC|STRING|STRCUR);
-			r->stfmt = -1;
+			if (value->val_type == AWK_STRNUM)
+				r->flags |= USER_INPUT;
+			r->stfmt = STFMT_UNUSED;
 			r->stptr = value->str_value.str;
 			r->stlen = value->str_value.len;
-			return true;
+#ifdef HAVE_MPFR
+			r->strndmode = MPFR_round_mode;
+#endif
+			return awk_true;
 		}
 		break;
+
+	case AWK_REGEX:
 	case AWK_UNDEFINED:
 	case AWK_SCALAR:
 	case AWK_VALUE_COOKIE:
 		break;
 
+
 	default:	/* AWK_ARRAY or invalid type */
-		return false;
+		return awk_false;
 	}
 
 	/* do it the hard (slow) way */
 	unref(node->var_value);
 	node->var_value = awk_value_to_node(value);
-	return true;
+	return awk_true;
 }
 
 /*
@@ -694,13 +1017,15 @@ api_sym_update_scalar(awk_ext_id_t id,
  * Any scalar value is fine, so only AWK_ARRAY (or an invalid type) is illegal.
  */
 
-static inline int
+static inline bool
 valid_subscript_type(awk_valtype_t valtype)
 {
 	switch (valtype) {
 	case AWK_UNDEFINED:
 	case AWK_NUMBER:
+	case AWK_STRNUM:
 	case AWK_STRING:
+	case AWK_REGEX:
 	case AWK_SCALAR:
 	case AWK_VALUE_COOKIE:
 		return true;
@@ -733,21 +1058,21 @@ api_get_array_element(awk_ext_id_t id,
 	    || result == NULL
 	    || index == NULL
 	    || ! valid_subscript_type(index->val_type))
-		return false;
+		return awk_false;
 
 	subscript = awk_value_to_node(index);
 
 	/* if it doesn't exist, return false */
 	if (in_array(array, subscript) == NULL) {
 		unref(subscript);
-		return false;
+		return awk_false;
 	}
 
 	aptr = assoc_lookup(array, subscript);
 
 	if (aptr == NULL) {	/* can't happen */
 		unref(subscript);
-		return false;
+		return awk_false;
 	}
 
 	unref(subscript);
@@ -768,7 +1093,6 @@ api_set_array_element(awk_ext_id_t id, awk_array_t a_cookie,
 	NODE *array = (NODE *)a_cookie;
 	NODE *tmp;
 	NODE *elem;
-	NODE **aptr;
 
 	/* don't check for index len zero, null str is ok as index */
 	if (   array == NULL
@@ -777,25 +1101,22 @@ api_set_array_element(awk_ext_id_t id, awk_array_t a_cookie,
 	    || index == NULL
 	    || value == NULL
 	    || ! valid_subscript_type(index->val_type))
-		return false;
+		return awk_false;
 
 	tmp = awk_value_to_node(index);
-	aptr = assoc_lookup(array, tmp);
-	unref(tmp);
-	unref(*aptr);
-	elem = *aptr = awk_value_to_node(value);
+	elem = awk_value_to_node(value);
 	if (elem->type == Node_var_array) {
 		elem->parent_array = array;
 		elem->vname = estrdup(index->str_value.str,
 					index->str_value.len);
-		make_aname(elem);
 	}
+	assoc_set(array, tmp, elem);
 
-	return true;
+	return awk_true;
 }
 
 /*
- * remove_element --- remove an array element 
+ * remove_element --- remove an array element
  *		common code used by multiple functions
  */
 
@@ -843,13 +1164,13 @@ api_del_array_element(awk_ext_id_t id,
 	    || (array->flags & NO_EXT_SET) != 0
 	    || index == NULL
 	    || ! valid_subscript_type(index->val_type))
-		return false;
+		return awk_false;
 
 	sub = awk_value_to_node(index);
 	remove_element(array, sub);
 	unref(sub);
 
-	return true;
+	return awk_true;
 }
 
 /*
@@ -864,10 +1185,10 @@ api_get_element_count(awk_ext_id_t id,
 	NODE *node = (NODE *) a_cookie;
 
 	if (count == NULL || node == NULL || node->type != Node_var_array)
-		return false;
+		return awk_false;
 
 	*count = node->table_size;
-	return true;
+	return awk_true;
 }
 
 /* api_create_array --- create a new array cookie to which elements may be added */
@@ -894,18 +1215,30 @@ api_clear_array(awk_ext_id_t id, awk_array_t a_cookie)
 	if (   node == NULL
 	    || node->type != Node_var_array
 	    || (node->flags & NO_EXT_SET) != 0)
-		return false;
+		return awk_false;
 
 	assoc_clear(node);
-	return true;
+	return awk_true;
 }
 
-/* api_flatten_array --- flatten out an array so that it can be looped over easily. */
+/* api_destroy_array --- destroy an array */
 
 static awk_bool_t
-api_flatten_array(awk_ext_id_t id,
+api_destroy_array(awk_ext_id_t id, awk_array_t a_cookie)
+{
+	if (! api_clear_array(id, a_cookie))
+		return awk_false;
+	freenode((NODE *) a_cookie);
+	return awk_true;
+}
+
+/* api_flatten_array_typed --- flatten out an array so that it can be looped over easily. */
+
+static awk_bool_t
+api_flatten_array_typed(awk_ext_id_t id,
 		awk_array_t a_cookie,
-		awk_flat_array_t **data)
+		awk_flat_array_t **data,
+		awk_valtype_t index_type, awk_valtype_t value_type)
 {
 	NODE **list;
 	size_t i, j;
@@ -914,16 +1247,15 @@ api_flatten_array(awk_ext_id_t id,
 
 	if (   array == NULL
 	    || array->type != Node_var_array
-	    || array->table_size == 0
+	    || assoc_empty(array)
 	    || data == NULL)
-		return false;
+		return awk_false;
 
 	alloc_size = sizeof(awk_flat_array_t) +
 			(array->table_size - 1) * sizeof(awk_element_t);
 
-	emalloc(*data, awk_flat_array_t *, alloc_size,
-			"api_flatten_array");
-	memset(*data, 0, alloc_size);
+	ezalloc(*data, awk_flat_array_t *, alloc_size,
+			"api_flatten_array_typed");
 
 	list = assoc_list(array, "@unsorted", ASORTI);
 
@@ -934,22 +1266,22 @@ api_flatten_array(awk_ext_id_t id,
 	for (i = j = 0; i < 2 * array->table_size; i += 2, j++) {
 		NODE *index, *value;
 
-		index = force_string(list[i]);
+		index = list[i];
 		value = list[i + 1]; /* number or string or subarray */
 
-		/* convert index and value to ext types */
+		/* Convert index and value to API types. */
 		if (! node_to_awk_value(index,
-				& (*data)->elements[j].index, AWK_UNDEFINED)) {
-			fatal(_("api_flatten_array: could not convert index %d\n"),
-						(int) i);
+				& (*data)->elements[j].index, index_type)) {
+			fatal(_("api_flatten_array_typed: could not convert index %d to %s"),
+						(int) i, valtype2str(index_type));
 		}
 		if (! node_to_awk_value(value,
-				& (*data)->elements[j].value, AWK_UNDEFINED)) {
-			fatal(_("api_flatten_array: could not convert value %d\n"),
-						(int) i);
+				& (*data)->elements[j].value, value_type)) {
+			fatal(_("api_flatten_array_typed: could not convert value %d to %s"),
+						(int) i, valtype2str(value_type));
 		}
 	}
-	return true;
+	return awk_true;
 }
 
 /*
@@ -963,7 +1295,7 @@ api_release_flattened_array(awk_ext_id_t id,
 		awk_array_t a_cookie,
 		awk_flat_array_t *data)
 {
-	NODE *array = a_cookie;
+	NODE *array = (NODE *) a_cookie;
 	NODE **list;
 	size_t i, j, k;
 
@@ -973,7 +1305,7 @@ api_release_flattened_array(awk_ext_id_t id,
 	    || array != (NODE *) data->opaque1
 	    || data->count != array->table_size
 	    || data->opaque2 == NULL)
-		return false;
+		return awk_false;
 
 	list = (NODE **) data->opaque2;
 
@@ -990,7 +1322,7 @@ api_release_flattened_array(awk_ext_id_t id,
 	efree(list);
 	efree(data);
 
-	return true;
+	return awk_true;
 }
 
 /* api_create_value --- create a cached value */
@@ -1000,18 +1332,20 @@ api_create_value(awk_ext_id_t id, awk_value_t *value,
 		awk_value_cookie_t *result)
 {
 	if (value == NULL || result == NULL)
-		return false;
+		return awk_false;
 
 	switch (value->val_type) {
 	case AWK_NUMBER:
+	case AWK_STRNUM:
 	case AWK_STRING:
+	case AWK_REGEX:
 		break;
 	default:
 		/* reject anything other than a simple scalar */
-		return false;
+		return awk_false;
 	}
 
-	return (*result = awk_value_to_node(value)) != NULL;
+	return (awk_bool_t) ((*result = awk_value_to_node(value)) != NULL);
 }
 
 /* api_release_value --- release a cached value */
@@ -1022,10 +1356,136 @@ api_release_value(awk_ext_id_t id, awk_value_cookie_t value)
 	NODE *val = (NODE *) value;
 
 	if (val == NULL)
-		return false;
+		return awk_false;
 
 	unref(val);
-	return true;
+	return awk_true;
+}
+
+/* api_get_mpfr --- allocate an mpfr_ptr */
+
+static void *
+api_get_mpfr(awk_ext_id_t id)
+{
+#ifdef HAVE_MPFR
+	mpfr_ptr p;
+	emalloc(p, mpfr_ptr, sizeof(mpfr_t), "api_get_mpfr");
+	mpfr_init(p);
+	return p;
+#else
+	fatal(_("api_get_mpfr: MPFR not supported"));
+	return NULL;	// silence compiler warning
+#endif
+}
+
+/* api_get_mpz --- allocate an mpz_ptr */
+
+static void *
+api_get_mpz(awk_ext_id_t id)
+{
+#ifdef HAVE_MPFR
+	mpz_ptr p;
+	emalloc(p, mpz_ptr, sizeof (mpz_t), "api_get_mpz");
+
+	mpz_init(p);
+	return p;
+#else
+	fatal(_("api_get_mpfr: MPFR not supported"));
+	return NULL;	// silence compiler warning
+#endif
+}
+
+/* api_get_file --- return a handle to an existing or newly opened file */
+
+static awk_bool_t
+api_get_file(awk_ext_id_t id, const char *name, size_t namelen, const char *filetype,
+		int fd, const awk_input_buf_t **ibufp, const awk_output_buf_t **obufp)
+{
+	const struct redirect *f;
+	int flag;	/* not used, sigh */
+	enum redirval redirtype;
+
+	if (name == NULL || namelen == 0) {
+		if (curfile == NULL) {
+			INSTRUCTION *pc;
+			int save_rule;
+			char *save_source;
+
+			if (nextfile(& curfile, false) <= 0)
+				return awk_false;
+
+			pc = main_beginfile;
+			/* save execution state */
+			save_rule = currule;
+			save_source = source;
+
+			for (;;) {
+				if (pc == NULL)
+					fatal(_("cannot find end of BEGINFILE rule"));
+				if (pc->opcode == Op_after_beginfile)
+					break;
+				pc = pc->nexti;
+			}
+			pc->opcode = Op_stop;
+		        (void) (*interpret)(main_beginfile);
+			pc->opcode = Op_after_beginfile;
+			after_beginfile(& curfile);
+			/* restore execution state */
+			currule = save_rule;
+			source = save_source;
+		}
+		*ibufp = &curfile->public;
+		*obufp = NULL;
+
+		return awk_true;
+	}
+
+	redirtype = redirect_none;
+	switch (filetype[0]) {
+	case '<':
+		if (filetype[1] == '\0')
+			redirtype = redirect_input;
+		break;
+	case '>':
+		switch (filetype[1]) {
+		case '\0':
+			redirtype = redirect_output;
+			break;
+		case '>':
+			if (filetype[2] == '\0')
+				redirtype = redirect_append;
+			break;
+		}
+		break;
+	case '|':
+		if (filetype[2] == '\0') {
+			switch (filetype[1]) {
+			case '>':
+				redirtype = redirect_pipe;
+				break;
+			case '<':
+				redirtype = redirect_pipein;
+				break;
+			case '&':
+				redirtype = redirect_twoway;
+				break;
+			}
+		}
+		break;
+	}
+
+	if (redirtype == redirect_none) {
+		warning(_("cannot open unrecognized file type `%s' for `%s'"),
+			filetype, name);
+		return awk_false;
+	}
+
+	if ((f = redirect_string(name, namelen, 0, redirtype, &flag, fd, false)) == NULL)
+		return awk_false;
+
+	*ibufp = f->iop ? & f->iop->public : NULL;
+	*obufp = f->output.fp ? & f->output : NULL;
+	return awk_true;
 }
 
 /*
@@ -1062,6 +1522,16 @@ gawk_api_t api_impl = {
 	/* data */
 	GAWK_API_MAJOR_VERSION,	/* major and minor versions */
 	GAWK_API_MINOR_VERSION,
+
+#ifdef HAVE_MPFR
+	__GNU_MP_VERSION,
+	__GNU_MP_VERSION_MINOR,
+	MPFR_VERSION_MAJOR,
+	MPFR_VERSION_MINOR,
+#else
+	0, 0, 0, 0,
+#endif
+
 	{ 0 },			/* do_flags */
 
 	/* registration functions */
@@ -1076,6 +1546,7 @@ gawk_api_t api_impl = {
 	api_fatal,
 	api_warning,
 	api_lintwarn,
+	api_nonfatal,
 
 	/* updating ERRNO */
 	api_update_ERRNO_int,
@@ -1105,8 +1576,22 @@ gawk_api_t api_impl = {
 	api_del_array_element,
 	api_create_array,
 	api_clear_array,
-	api_flatten_array,
+	api_flatten_array_typed,
 	api_release_flattened_array,
+
+	/* Memory allocation */
+	malloc,
+	calloc,
+	realloc,
+	free,
+	api_get_mpfr,
+	api_get_mpz,
+
+	/* Find/open a file */
+	api_get_file,
+
+	/* Additional array hook to destroy an array */
+	api_destroy_array,
 };
 
 /* init_ext_api --- init the extension API */
@@ -1115,12 +1600,13 @@ void
 init_ext_api()
 {
 	/* force values to 1 / 0 */
-	api_impl.do_flags[0] = (do_lint ? 1 : 0);
-	api_impl.do_flags[1] = (do_traditional ? 1 : 0);
-	api_impl.do_flags[2] = (do_profile ? 1 : 0);
-	api_impl.do_flags[3] = (do_sandbox ? 1 : 0);
-	api_impl.do_flags[4] = (do_debug ? 1 : 0);
-	api_impl.do_flags[5] = (do_mpfr ? 1 : 0);
+	api_impl.do_flags[gawk_do_lint] = (do_lint ? 1 : 0);
+	api_impl.do_flags[gawk_do_traditional] = (do_traditional ? 1 : 0);
+	api_impl.do_flags[gawk_do_profile] = (do_profile ? 1 : 0);
+	api_impl.do_flags[gawk_do_sandbox] = (do_sandbox ? 1 : 0);
+	api_impl.do_flags[gawk_do_debug] = (do_debug ? 1 : 0);
+	api_impl.do_flags[gawk_do_mpfr] = (do_mpfr ? 1 : 0);
+	api_impl.do_flags[gawk_do_csv] = (do_csv ? 1 : 0);
 }
 
 /* update_ext_api --- update the variables in the API that can change */
@@ -1141,3 +1627,58 @@ print_ext_versions(void)
 	for (p = vi_head; p != NULL; p = p->next)
 		printf("%s\n", p->version);
 }
+
+/* valtype2str --- return a printable representation of a value type */
+
+static const char *
+valtype2str(awk_valtype_t type)
+{
+	static char buf[100];
+
+	// Important: keep in same order as in gawkapi.h!
+	static const char *values[] = {
+		"AWK_UNDEFINED",
+		"AWK_NUMBER",
+		"AWK_STRING",
+		"AWK_REGEX",
+		"AWK_STRNUM",
+		"AWK_ARRAY",
+		"AWK_SCALAR",
+		"AWK_VALUE_COOKIE",
+	};
+
+	if (AWK_UNDEFINED <= type && type <= AWK_VALUE_COOKIE)
+		return values[(int) type];
+
+	sprintf(buf, "unknown type! (%d)", (int) type);
+
+	return buf;
+}
+
+/* ns_lookup --- correctly build name before looking it up */
+
+static NODE *
+ns_lookup(const char *name_space, const char *name, char **fullname)
+{
+	assert(name_space != NULL);
+	assert(name != NULL);
+
+	if (name_space[0] == '\0' || strcmp(name_space, awk_namespace) == 0) {
+		if (fullname != NULL)
+			*fullname = estrdup(name, strlen(name));
+		return lookup(name);
+	}
+
+	size_t len = strlen(name_space) + 2 + strlen(name) + 1;
+	char *buf;
+	emalloc(buf, char *, len, "ns_lookup");
+	sprintf(buf, "%s::%s", name_space, name);
+
+	NODE *f = lookup(buf);
+	if (fullname != NULL)
+		*fullname = buf;
+	else
+		efree((void *) buf);
+
+	return f;
+}
