@@ -2,10 +2,10 @@
  * interpret.h ---  run a list of instructions.
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2022,
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2023,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  *
@@ -275,8 +275,7 @@ uninitialized_scalar:
 				PUSH(m);
 		 		break;
 			}
- 			/* else
-				fall through */
+ 			/* fall through */
 		case Op_push_array:
 			PUSH(pc->memory);
 			break;
@@ -344,7 +343,9 @@ uninitialized_scalar:
 				}
 			}
 
-			if (r->type == Node_val || r->type == Node_elem_new)
+			if (r->type == Node_val
+			    || r->type == Node_var
+			    || r->type == Node_elem_new)
 				UPREF(r);
 			PUSH(r);
 			break;
@@ -482,9 +483,10 @@ uninitialized_scalar:
 		case Op_lint_plus:
 			// no need to check do_lint, this opcode won't
 			// be generated if that's not true
-			t1 = TOP();
-			t2 = PEEK(1);
-			if ((t1->flags & STRING) != 0 && (t2->flags & STRING) != 0)
+			t1 = fixtype(TOP());
+			t2 = fixtype(PEEK(1));
+			if ((t1->flags & (STRING|USER_INPUT)) == STRING
+			    && (t2->flags & (STRING|USER_INPUT)) == STRING)
 				lintwarn(_("operator `+' used on two string values"));
 			break;
 
@@ -792,6 +794,7 @@ mod:
 			break;
 
 		case Op_store_field:
+		case Op_store_field_exp:
 		{
 			/* field assignment optimization,
 			 * see awkgram.y (optimize_assignment)
@@ -814,6 +817,10 @@ mod:
 			UNFIELD(*lhs, r);
 			/* field variables need the string representation: */
 			force_string(*lhs);
+			if (op == Op_store_field_exp) {
+				UPREF(*lhs);
+				PUSH(*lhs);
+			}
 		}
 			break;
 
